@@ -17,6 +17,7 @@ use crate::performer::Machine;
 use renderable::Cursor;
 use renderable::RenderableContent;
 use rio_backend::config::Shell;
+
 use rio_backend::crosswords::{Crosswords, MIN_COLUMNS, MIN_LINES};
 use rio_backend::error::{RioError, RioErrorLevel, RioErrorType};
 use rio_backend::event::EventListener;
@@ -59,21 +60,22 @@ impl<T: rio_backend::event::EventListener> Drop for Context<T> {
 impl<T: EventListener> Context<T> {
     #[inline]
     pub fn set_selection(&mut self, selection_range: Option<SelectionRange>) {
-        let has_updated = (self.renderable_content.selection_range.is_none()
-            && selection_range.is_some())
-            || (self.renderable_content.selection_range.is_some()
-                && selection_range.is_none());
-        self.renderable_content.selection_range = selection_range;
+        let has_updated = self.renderable_content.selection_range != selection_range;
 
         if has_updated {
-            self.renderable_content.has_pending_updates = true;
+            let mut terminal = self.terminal.lock();
+            let display_offset = terminal.display_offset();
+            terminal.update_selection_damage(selection_range, display_offset);
         }
+
+        self.renderable_content.selection_range = selection_range;
     }
 
     #[inline]
     pub fn set_hyperlink_range(&mut self, hyperlink_range: Option<SelectionRange>) {
         self.renderable_content.hyperlink_range = hyperlink_range;
-        self.renderable_content.has_pending_updates = true;
+        self.renderable_content.has_pending_updates =
+            Some(rio_backend::event::TerminalDamage::Full);
     }
 
     #[inline]
@@ -81,37 +83,6 @@ impl<T: EventListener> Context<T> {
         self.renderable_content.hyperlink_range.is_some()
     }
 
-    #[inline]
-    pub fn renderable_content(&mut self) -> &RenderableContent {
-        let mut has_ime = false;
-        if let Some(preedit) = self.ime.preedit() {
-            if let Some(content) = preedit.text.chars().next() {
-                self.renderable_content.cursor.content = content;
-                self.renderable_content.cursor.is_ime_enabled = true;
-                has_ime = true;
-            }
-        }
-
-        if !has_ime {
-            self.renderable_content.cursor.is_ime_enabled = false;
-            self.renderable_content.cursor.content =
-                self.renderable_content.cursor.content_ref;
-        }
-
-        let terminal = self.terminal.lock();
-        self.renderable_content.update(
-            terminal.visible_rows(),
-            terminal.display_offset(),
-            terminal.cursor(),
-            terminal.is_fully_damaged(),
-            terminal.colors,
-            terminal.blinking_cursor,
-        );
-        drop(terminal);
-
-        &self.renderable_content
-    }
-
     #[inline]
     pub fn cursor_from_ref(&self) -> Cursor {
         Cursor {
@@ -130,11 +101,12 @@ pub struct ContextManagerConfig {
     pub use_fork: bool,
     pub working_dir: Option<String>,
     pub spawn_performer: bool,
-    pub use_current_path: bool,
+    pub cwd: bool,
     pub is_native: bool,
     pub should_update_title_extra: bool,
     pub split_color: [f32; 4],
     pub title: rio_backend::config::title::Title,
+    pub keyboard: rio_backend::config::keyboard::Keyboard,
 }
 
 pub struct ContextManager<T: EventListener> {
@@ -203,7 +175,7 @@ pub fn create_mock_context<
         spawn_performer: false,
         is_native: false,
         should_update_title_extra: false,
-        use_current_path: false,
+        cwd: false,
         ..ContextManagerConfig::default()
     };
     ContextManager::create_context(
@@ -424,7 +396,7 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
             spawn_performer: false,
             is_native: false,
             should_update_title_extra: false,
-            use_current_path: false,
+            cwd: false,
             ..ContextManagerConfig::default()
         };
         let initial_context = ContextManager::create_context(
@@ -577,9 +549,11 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
             return;
         }
         self.switch_to_next();
-        // Make sure first split is selected
+        // Make sure first split is selected - get the root key
         let current_tab = &mut self.contexts[self.current_index];
-        current_tab.current = 0;
+        if let Some(root) = current_tab.root {
+            current_tab.current = root;
+        }
         self.current_route = self.current().route_id;
     }
 
@@ -590,12 +564,35 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
             return;
         }
         self.switch_to_prev();
-        // Make sure last split is selected
+        // Make sure last split is selected - get the last key in order
         let current_tab = &mut self.contexts[self.current_index];
-        current_tab.current = current_tab.len() - 1;
+        let ordered_keys = current_tab.get_ordered_keys();
+        if let Some(&last_key) = ordered_keys.last() {
+            current_tab.current = last_key;
+        }
         self.current_route = self.current().route_id;
     }
 
+    #[inline]
+    pub fn move_divider_up(&mut self, amount: f32) -> bool {
+        self.contexts[self.current_index].move_divider_up(amount)
+    }
+
+    #[inline]
+    pub fn move_divider_down(&mut self, amount: f32) -> bool {
+        self.contexts[self.current_index].move_divider_down(amount)
+    }
+
+    #[inline]
+    pub fn move_divider_left(&mut self, amount: f32) -> bool {
+        self.contexts[self.current_index].move_divider_left(amount)
+    }
+
+    #[inline]
+    pub fn move_divider_right(&mut self, amount: f32) -> bool {
+        self.contexts[self.current_index].move_divider_right(amount)
+    }
+
     #[inline]
     pub fn select_tab(&mut self, tab_index: usize) {
         if self.config.is_native {
@@ -659,8 +656,8 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
     }
 
     #[inline]
-    pub fn grid_objects(&self) -> Vec<Object> {
-        self.contexts[self.current_index].objects()
+    pub fn extend_with_grid_objects(&self, target: &mut Vec<Object>) {
+        self.contexts[self.current_index].extend_with_objects(target);
     }
 
     #[inline]
@@ -684,7 +681,7 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
                 self.event_proxy
                     .send_event(RioEvent::Title(content.to_owned()), self.window_id);
 
-                id.push_str(&format!("{}{};", i, content));
+                id.push_str(&format!("{i}{content};"));
 
                 if self.config.should_update_title_extra {
                     self.titles.set_key_val(
@@ -851,7 +848,7 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
 
     pub fn split(&mut self, rich_text_id: usize, split_down: bool) {
         let mut working_dir = self.config.working_dir.clone();
-        if self.config.use_current_path {
+        if self.config.cwd {
             #[cfg(not(target_os = "windows"))]
             {
                 let current_context = self.current();
@@ -920,7 +917,7 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
         );
 
         let context_manager_config = ContextManagerConfig {
-            use_current_path: config.navigation.use_current_path,
+            cwd: config.navigation.current_working_directory,
             shell,
             working_dir,
             spawn_performer: true,
@@ -932,6 +929,7 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
             should_update_title_extra: !config.navigation.color_automation.is_empty(),
             split_color: config.colors.split,
             title: config.title,
+            keyboard: config.keyboard,
         };
 
         self.acc_current_route += 1;
@@ -965,7 +963,7 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
     #[inline]
     pub fn add_context(&mut self, redirect: bool, rich_text_id: usize) {
         let mut working_dir = self.config.working_dir.clone();
-        if self.config.use_current_path {
+        if self.config.cwd {
             #[cfg(not(target_os = "windows"))]
             {
                 let current_context = self.current();
@@ -1005,13 +1003,20 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
             self.acc_current_route += 1;
             let current = self.current();
             let cursor = current.cursor_from_ref();
+            let mut dimension = current.dimension;
+
+            // If current has splits then shouldn't use that dimension
+            if self.current_grid().len() > 1 {
+                dimension = self.current_grid().grid_dimension();
+            }
+
             match ContextManager::create_context(
                 (&cursor, current.renderable_content.has_blinking_enabled),
                 self.event_proxy.clone(),
                 self.window_id,
                 self.acc_current_route,
                 rich_text_id,
-                self.current_grid().grid_dimension(),
+                dimension,
                 &cloned_config,
             ) {
                 Ok(new_context) => {
@@ -1280,42 +1285,42 @@ pub mod test {
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 1);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 1);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 2);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 2);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 2);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 3);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 3);
-        assert_eq!(context_manager.contexts[current_index].current, 1);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 1);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 4);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 0);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
     }
 
     #[test]
@@ -1343,42 +1348,42 @@ pub mod test {
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 4);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 3);
-        assert_eq!(context_manager.contexts[current_index].current, 1);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 1);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 3);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 2);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 2);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 2);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 1);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 1);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 0);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
     }
 
     #[test]
@@ -1404,43 +1409,43 @@ pub mod test {
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 1);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 1);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 2);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 2);
 
         context_manager.switch_to_next_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 2);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         // Prev
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 2);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 2);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 1);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 1);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 1);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
 
         context_manager.switch_to_prev_split_or_tab();
         current_index = context_manager.current_index;
         assert_eq!(current_index, 0);
-        assert_eq!(context_manager.contexts[current_index].current, 0);
+        assert_eq!(context_manager.contexts[current_index].current_index(), 0);
     }
 
     #[test]
