@@ -1,7 +1,12 @@
+mod char_cache;
+mod font_cache;
 pub mod navigation;
 mod search;
 pub mod utils;
 
+use crate::renderer::font_cache::FontCache;
+use char_cache::CharCache;
+
 use crate::ansi::CursorShape;
 use crate::context::renderable::{Cursor, RenderableContent};
 use crate::context::ContextManager;
@@ -26,7 +31,6 @@ use rio_backend::sugarloaf::{
 use std::collections::HashMap;
 use std::ops::RangeInclusive;
 
-use rustc_hash::FxHashMap;
 use unicode_width::UnicodeWidthChar;
 
 #[derive(Default)]
@@ -43,7 +47,7 @@ pub struct Renderer {
     pub colors: List,
     pub navigation: ScreenNavigation,
     unfocused_split_opacity: f32,
-    last_active: usize,
+    last_active: Option<slotmap::DefaultKey>,
     pub config_has_blinking_enabled: bool,
     pub config_blinking_interval: u64,
     ignore_selection_fg_color: bool,
@@ -56,10 +60,8 @@ pub struct Renderer {
     // the same r,g,b with the mutated alpha channel.
     pub dynamic_background: ([f32; 4], wgpu::Color, bool),
     font_context: rio_backend::sugarloaf::font::FontLibrary,
-    font_cache: FxHashMap<
-        (char, rio_backend::sugarloaf::font_introspector::Attributes),
-        (usize, f32),
-    >,
+    font_cache: FontCache,
+    char_cache: CharCache,
 }
 
 impl Renderer {
@@ -90,9 +92,9 @@ impl Renderer {
                 .insert(rule.path.clone(), rule.color);
         }
 
-        Renderer {
+        let mut renderer = Renderer {
             unfocused_split_opacity: config.navigation.unfocused_split_opacity,
-            last_active: 0,
+            last_active: None,
             use_drawable_chars: config.fonts.use_drawable_chars,
             draw_bold_text_with_light_colors: config.draw_bold_text_with_light_colors,
             macos_use_unified_titlebar: config.window.macos_use_unified_titlebar,
@@ -110,9 +112,15 @@ impl Renderer {
             named_colors,
             dynamic_background,
             search: Search::default(),
-            font_cache: FxHashMap::default(),
+            font_cache: FontCache::new(),
             font_context: font_context.clone(),
-        }
+            char_cache: CharCache::new(),
+        };
+
+        // Pre-populate font cache with common characters for better performance
+        renderer.font_cache.pre_populate(font_context);
+
+        renderer
     }
 
     #[inline]
@@ -188,8 +196,6 @@ impl Renderer {
 
         if square.flags.contains(Flags::UNDERLINE) {
             decoration = Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                offset: -1.0,
-                size: 1.0,
                 is_doubled: false,
                 shape: UnderlineShape::Regular,
             }));
@@ -197,29 +203,21 @@ impl Renderer {
             decoration = Some(FragmentStyleDecoration::Strikethrough);
         } else if square.flags.contains(Flags::DOUBLE_UNDERLINE) {
             decoration = Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                offset: -1.0,
-                size: 1.0,
                 is_doubled: true,
                 shape: UnderlineShape::Regular,
             }));
         } else if square.flags.contains(Flags::DOTTED_UNDERLINE) {
             decoration = Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                offset: -1.0,
-                size: 2.0,
                 is_doubled: false,
                 shape: UnderlineShape::Dotted,
             }));
         } else if square.flags.contains(Flags::DASHED_UNDERLINE) {
             decoration = Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                offset: -1.0,
-                size: 2.0,
                 is_doubled: false,
                 shape: UnderlineShape::Dashed,
             }));
         } else if square.flags.contains(Flags::UNDERCURL) {
             decoration = Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                offset: -1.0,
-                size: 2.0,
                 is_doubled: false,
                 shape: UnderlineShape::Curly,
             }));
@@ -255,10 +253,15 @@ impl Renderer {
         let hyperlink_range = renderable_content.hyperlink_range;
         let selection_range = renderable_content.selection_range;
         let columns: usize = row.len();
-        let mut content = String::default();
+        let mut content = String::with_capacity(columns);
         let mut last_char_was_space = false;
         let mut last_style = FragmentStyle::default();
 
+        // Collect all characters that need font lookups to batch them
+        let mut font_lookups = Vec::new();
+        let mut styles_and_chars = Vec::with_capacity(columns);
+
+        // First pass: collect all styles and identify font cache misses
         for column in 0..columns {
             let square = &row.inner[column];
 
@@ -281,8 +284,6 @@ impl Renderer {
             {
                 style.decoration =
                     Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                        offset: -1.0,
-                        size: -1.0,
                         is_doubled: false,
                         shape: UnderlineShape::Regular,
                     }));
@@ -324,17 +325,6 @@ impl Renderer {
             }
 
             if square.flags.contains(Flags::GRAPHICS) {
-                // let graphics = square.graphics().map(|graphics| {
-                //     graphics
-                //         .iter()
-                //         .map(|graphic| Graphic {
-                //             id: graphic.texture.id,
-                //             offset_x: graphic.offset_x,
-                //             offset_y: graphic.offset_y,
-                //         })
-                //         .collect::<_>()
-                // });
-                // style.media = Some(graphics);
                 let graphic = &square.graphics().unwrap()[0];
                 style.media = Some(Graphic {
                     id: graphic.texture.id,
@@ -344,28 +334,10 @@ impl Renderer {
                 style.background_color = None;
             }
 
-            // In case we have a drawable character then we will:
-            // 1. Ignore font cache and find width since it's known already.
-            // 2. Wrap up the content and send to sugarloaf.
-            //
-            // TODO: In the future it should use same logic to render everything
-            // at once.
+            // Handle drawable characters
             if self.use_drawable_chars {
                 if let Some(character) = drawable_character(square_content) {
                     style.drawable_char = Some(character);
-                    if !content.is_empty() {
-                        if let Some(line) = line_opt {
-                            builder.add_text_on_line(line, &content, last_style);
-                        } else {
-                            builder.add_text(&content, last_style);
-                        }
-                        content.clear();
-                    }
-
-                    last_style = style;
-
-                    // Ignore font shaping
-                    content.push(' ');
                 }
             }
 
@@ -377,34 +349,56 @@ impl Renderer {
                     style.font_id = *font_id;
                     style.width = *width;
                 } else {
-                    let mut width = square.c.width().unwrap_or(1) as f32;
-                    let mut font_ctx = self.font_context.inner.lock();
+                    // Mark this character for font lookup
+                    font_lookups.push((
+                        styles_and_chars.len(),
+                        square_content,
+                        style.font_attrs,
+                    ));
+                }
+            }
 
-                    // There is no simple way to define what's emoji
-                    // could have to refer to the Unicode tables. However it could
-                    // be leading to misleading results. For example if we used
-                    // unicode and internationalization functionalities like
-                    // https://github.com/open-i18n/rust-unic/, then characters
-                    // like "◼" would be valid emojis. For a terminal context,
-                    // the character "◼" is not an emoji and should be treated as
-                    // single width. So, we completely rely on what font is
-                    // being used and then set width 2 for it.
-                    if let Some((font_id, is_emoji)) =
-                        font_ctx.find_best_font_match(square_content, &style)
-                    {
-                        style.font_id = font_id;
-                        if is_emoji {
-                            width = 2.0;
-                        }
+            styles_and_chars.push((style, square_content, column));
+        }
+
+        // Batch font lookups with a single lock acquisition
+        if !font_lookups.is_empty() {
+            let font_ctx = self.font_context.inner.read();
+            for (style_index, square_content, font_attrs) in font_lookups {
+                let mut width = square_content.width().unwrap_or(1) as f32;
+                let style = &mut styles_and_chars[style_index].0;
+
+                if let Some((font_id, is_emoji)) =
+                    font_ctx.find_best_font_match(square_content, style)
+                {
+                    style.font_id = font_id;
+                    if is_emoji {
+                        width = 2.0;
                     }
-                    style.width = width;
+                }
+                style.width = width;
 
-                    self.font_cache.insert(
-                        (square_content, style.font_attrs),
-                        (style.font_id, style.width),
-                    );
-                };
+                self.font_cache
+                    .insert((square_content, font_attrs), (style.font_id, style.width));
+            }
+        }
 
+        // Second pass: render the line using the resolved styles
+        for (style, square_content, column) in styles_and_chars {
+            // Handle drawable characters
+            if style.drawable_char.is_some() {
+                if !content.is_empty() {
+                    if let Some(line) = line_opt {
+                        builder.add_text_on_line(line, &content, last_style);
+                    } else {
+                        builder.add_text(&content, last_style);
+                    }
+                    content.clear();
+                }
+
+                last_style = style;
+                content.push(' '); // Ignore font shaping
+            } else {
                 if square_content == ' ' {
                     if !last_char_was_space {
                         if !content.is_empty() {
@@ -635,8 +629,6 @@ impl Renderer {
             CursorShape::Underline => {
                 style.decoration =
                     Some(FragmentStyleDecoration::Underline(UnderlineInfo {
-                        offset: 0.0,
-                        size: 3.0,
                         is_doubled: false,
                         shape: UnderlineShape::Regular,
                     }));
@@ -693,40 +685,56 @@ impl Renderer {
                         )
                         .build();
                 } else {
-                    let mut style = FragmentStyle {
+                    let style = FragmentStyle {
                         color: self.named_colors.foreground,
                         ..FragmentStyle::default()
                     };
                     let line = content.sel(search_rich_text);
                     line.clear().new_line().add_text("Search: ", style);
 
+                    // Collect characters that need font lookups
+                    let mut font_lookups = Vec::new();
+                    let mut char_styles = Vec::new();
+
                     for character in active_search_content.chars() {
+                        let mut char_style = style;
                         if let Some((font_id, width)) =
                             self.font_cache.get(&(character, style.font_attrs))
                         {
-                            style.font_id = *font_id;
-                            style.width = *width;
+                            char_style.font_id = *font_id;
+                            char_style.width = *width;
                         } else {
+                            font_lookups.push((char_styles.len(), character));
+                        }
+                        char_styles.push((char_style, character));
+                    }
+
+                    // Batch font lookups with a single lock acquisition
+                    if !font_lookups.is_empty() {
+                        let font_ctx = self.font_context.inner.read();
+                        for (style_index, character) in font_lookups {
                             let mut width = character.width().unwrap_or(1) as f32;
-                            let mut font_ctx = self.font_context.inner.lock();
+                            let char_style = &mut char_styles[style_index].0;
 
-                            // Note we don't update cache from search bar
                             if let Some((font_id, is_emoji)) =
-                                font_ctx.find_best_font_match(character, &style)
+                                font_ctx.find_best_font_match(character, char_style)
                             {
-                                style.font_id = font_id;
+                                char_style.font_id = font_id;
                                 if is_emoji {
                                     width = 2.0;
                                 }
                             }
-                            style.width = width;
-                        };
+                            char_style.width = width;
+                        }
+                    }
 
+                    // Render all characters
+                    for (char_style, character) in char_styles {
                         line.add_text_on_line(
                             // Add on first line
                             1,
-                            &character.to_string(),
-                            style,
+                            self.char_cache.get_str(character),
+                            char_style,
                         );
                     }
 
@@ -744,6 +752,8 @@ impl Renderer {
         hints: &mut Option<HintMatches>,
         focused_match: &Option<RangeInclusive<Pos>>,
     ) {
+        // let start = std::time::Instant::now();
+
         // In case rich text for search was not created
         let has_search = self.search.active_search.is_some();
         if has_search && self.search.rich_text_id.is_none() {
@@ -754,17 +764,18 @@ impl Renderer {
 
         let mut graphic_queues: Option<Vec<UpdateQueues>> = None;
 
+        // let grid_start = std::time::Instant::now();
         let grid = context_manager.current_grid_mut();
-        let active_index = grid.current;
+        let active_key = grid.current;
         let mut has_active_changed = false;
-        if self.last_active != active_index {
+        if self.last_active != Some(active_key) {
             has_active_changed = true;
 
-            self.last_active = active_index;
+            self.last_active = Some(active_key);
         }
 
-        for (index, grid_context) in grid.contexts_mut().iter_mut().enumerate() {
-            let is_active = active_index == index;
+        for (key, grid_context) in grid.contexts_mut().iter_mut() {
+            let is_active = active_key == key;
             let context = grid_context.context_mut();
 
             let mut has_ime = false;
@@ -785,13 +796,11 @@ impl Renderer {
             // let duration = start.elapsed();
             // println!("Time elapsed in antes is: {:?}", duration);
             // let renderable_content = context.renderable_content();
-            let force_full_damage = has_active_changed
-                || context.renderable_content.has_pending_updates
-                || is_active
-                    && (context.renderable_content.selection_range.is_some()
-                        || hints.is_some());
+            let force_full_damage = has_active_changed || is_active && hints.is_some();
 
             let mut specific_lines = None;
+            // let duration = start.elapsed();
+            // println!("Time elapsed in before lock is: {:?}", duration);
             let (colors, display_offset, blinking_cursor, visible_rows) = {
                 let mut terminal = context.terminal.lock();
                 let result = (
@@ -803,6 +812,12 @@ impl Renderer {
 
                 context.renderable_content.cursor.state = terminal.cursor();
 
+                // Early exit if no rendering is needed
+                if !force_full_damage && !terminal.needs_render() {
+                    terminal.reset_damage();
+                    return;
+                }
+
                 if let Some(queues_to_add) = terminal.graphics_take_queues() {
                     if let Some(ref mut queues) = graphic_queues {
                         queues.push(queues_to_add);
@@ -811,22 +826,37 @@ impl Renderer {
                     }
                 }
 
+                // Check for partial damage to optimize rendering
                 if !force_full_damage && !terminal.is_fully_damaged() {
                     if let TermDamage::Partial(lines) = terminal.damage() {
+                        // Pre-allocate with estimated capacity to reduce allocations
+                        let capacity =
+                            lines.size_hint().1.unwrap_or(lines.size_hint().0).min(256);
                         let mut own_lines =
-                            std::collections::HashSet::with_capacity(result.3.len());
+                            std::collections::HashSet::with_capacity(capacity + 1);
                         for line in lines {
                             own_lines.insert(line.line);
                         }
-                        specific_lines = Some(own_lines);
+                        // Only set specific_lines if there are actually damaged lines
+                        if !own_lines.is_empty() {
+                            specific_lines = Some(own_lines);
+                        }
                     };
                 }
                 terminal.reset_damage();
                 result
             };
 
-            // let duration = start.elapsed();
-            // println!("Time elapsed in antes-antes is: {:?}", duration);
+            // If the last line is bigger than the actual visible rows, then some resize
+            // has happened. In this case, request full draw.
+            if let Some(ref lines) = specific_lines {
+                if let Some(max_value) = lines.iter().max() {
+                    if max_value > &(visible_rows.len() - 1) {
+                        specific_lines = None;
+                    }
+                }
+            }
+
             let rich_text_id = context.rich_text_id;
 
             let mut is_cursor_visible =
@@ -845,17 +875,41 @@ impl Renderer {
                     }
 
                     if should_blink {
-                        context.renderable_content.is_blinking_cursor_visible =
-                            !context.renderable_content.is_blinking_cursor_visible;
-                        if let Some(ref mut lines) = specific_lines {
-                            lines.insert(
-                                context.renderable_content.cursor.state.pos.row.0
-                                    as usize,
-                            );
+                        let now = std::time::Instant::now();
+                        let should_toggle = if let Some(last_blink) =
+                            context.renderable_content.last_blink_toggle
+                        {
+                            now.duration_since(last_blink).as_millis()
+                                >= self.config_blinking_interval as u128
+                        } else {
+                            // First time: start with cursor visible and set initial timing
+                            context.renderable_content.is_blinking_cursor_visible = true;
+                            context.renderable_content.last_blink_toggle = Some(now);
+                            false // Don't toggle on first frame
+                        };
+
+                        if should_toggle {
+                            context.renderable_content.is_blinking_cursor_visible =
+                                !context.renderable_content.is_blinking_cursor_visible;
+                            context.renderable_content.last_blink_toggle = Some(now);
+
+                            if let Some(ref mut lines) = specific_lines {
+                                lines.insert(
+                                    context.renderable_content.cursor.state.pos.row.0
+                                        as usize,
+                                );
+                            }
                         }
                     } else {
+                        // When not blinking (e.g., during typing), ensure cursor is visible
                         context.renderable_content.is_blinking_cursor_visible = true;
+                        // Reset blink timing when not blinking so it starts fresh when blinking resumes
+                        context.renderable_content.last_blink_toggle = None;
                     }
+                } else {
+                    // When there's a selection, keep cursor visible and reset blink timing
+                    context.renderable_content.is_blinking_cursor_visible = true;
+                    context.renderable_content.last_blink_toggle = None;
                 }
 
                 is_cursor_visible = context.renderable_content.is_blinking_cursor_visible;
@@ -865,10 +919,10 @@ impl Renderer {
                 is_cursor_visible = true;
             }
 
-            context.renderable_content.has_pending_updates = false;
             let content = sugarloaf.content();
             match specific_lines {
                 None => {
+                    // let start = std::time::Instant::now();
                     content.sel(rich_text_id);
                     content.clear();
                     for (i, row) in visible_rows.iter().enumerate() {
@@ -889,7 +943,10 @@ impl Renderer {
                     }
                     content.build();
                     // let duration = start.elapsed();
-                    // println!("Time elapsed in -renderer.TermDamage::Full is: {:?}", duration);
+                    // println!(
+                    //     "Time elapsed in -renderer.TermDamage::Full is: {:?}",
+                    //     duration
+                    // );
                 }
                 Some(lines) => {
                     content.sel(rich_text_id);
@@ -897,19 +954,27 @@ impl Renderer {
                         let has_cursor = is_cursor_visible
                             && context.renderable_content.cursor.state.pos.row == line;
                         content.clear_line(line);
-                        self.create_line(
-                            content,
-                            &visible_rows[line],
-                            has_cursor,
-                            Some(line),
-                            Line(line as i32),
-                            &context.renderable_content,
-                            hints,
-                            focused_match,
-                            &colors,
-                            is_active,
-                        );
+                        if let Some(visible_row) = visible_rows.get(line) {
+                            self.create_line(
+                                content,
+                                visible_row,
+                                has_cursor,
+                                Some(line),
+                                Line((line as i32) - display_offset as i32),
+                                &context.renderable_content,
+                                hints,
+                                focused_match,
+                                &colors,
+                                is_active,
+                            );
+                        }
                     }
+
+                    // let duration = start.elapsed();
+                    // println!(
+                    //     "Time elapsed in -renderer.TermDamage::Lines is: {:?}",
+                    //     duration
+                    // );
                 }
             };
         }
@@ -954,10 +1019,21 @@ impl Renderer {
             self.search.rich_text_id = None;
         }
 
+        // let duration = start.elapsed();
+        // println!(
+        //     "Time elapsed before extend_with_grid_objects is: {:?}",
+        //     duration
+        // );
         context_manager.extend_with_grid_objects(&mut objects);
+        // let duration = start.elapsed();
+        // println!(
+        //     "Time elapsed after extend_with_grid_objects is: {:?}",
+        //     duration
+        // );
         sugarloaf.set_objects(objects);
 
         sugarloaf.render();
+
         // let duration = start.elapsed();
         // println!("Time elapsed in -renderer.update() is: {:?}", duration);
     }
