@@ -1,5 +1,6 @@
 use serde::{Deserialize, Serialize};
 use std::fmt::Display;
+use sugarloaf::Filter;
 
 #[derive(Debug, Clone, PartialEq, Deserialize, Serialize)]
 pub struct Renderer {
@@ -9,8 +10,40 @@ pub struct Renderer {
     pub backend: Backend,
     #[serde(default = "bool::default", rename = "disable-unfocused-render")]
     pub disable_unfocused_render: bool,
-    #[serde(default = "Option::default", rename = "target-fps")]
-    pub target_fps: Option<u64>,
+    #[serde(
+        default = "default_disable_occluded_render",
+        rename = "disable-occluded-render"
+    )]
+    pub disable_occluded_render: bool,
+    #[serde(default = "Vec::default")]
+    pub filters: Vec<Filter>,
+    #[serde(default = "RendererStategy::default")]
+    pub strategy: RendererStategy,
+}
+
+fn default_disable_occluded_render() -> bool {
+    true
+}
+
+#[derive(Default, Debug, Clone, PartialEq, Deserialize, Serialize)]
+pub enum RendererStategy {
+    #[default]
+    #[serde(alias = "events")]
+    Events,
+    #[serde(alias = "game")]
+    Game,
+}
+
+impl RendererStategy {
+    #[inline]
+    pub fn is_game(&self) -> bool {
+        self == &RendererStategy::Game
+    }
+
+    #[inline]
+    pub fn is_event_based(&self) -> bool {
+        self == &RendererStategy::Events
+    }
 }
 
 #[allow(clippy::derivable_impls)]
@@ -20,7 +53,9 @@ impl Default for Renderer {
             performance: Performance::default(),
             backend: Backend::default(),
             disable_unfocused_render: false,
-            target_fps: None,
+            disable_occluded_render: default_disable_occluded_render(),
+            filters: Vec::default(),
+            strategy: RendererStategy::Events,
         }
     }
 }
